/* Todo:
 *   x add addrule to save
 *   x add rules to show
 *   - finish editrule
 *   x finish moverule
 *   - code to recalc top and bottom priorities when rule deleted
 *   - isolate formula resolver
 *   - add rule parser to table builder
 *   - code to work out next rule to apply
 *   x fix to make sure rule priorities dont conflict?
 *
 *
 *
 *
 */



#include <stdio.h>
#include <malloc.h>
#include "returnValues.h"
#include "intelhex.h"
#include "rules.h"
#include <unistd.h>
#include <string.h>
#include "eq5.h"
#include "version.h"
#include "danStuff.h"


// -lreadline
#include <readline/readline.h>
#include <readline/history.h>



//#define debug(k)   printf("%s\n", (k));
#define debug(k)  /* (k) */

#define REPLYSZ       512


/*

default n           set unspecified memory locatoins to n
save    filename    save project as filename
load    filename    load project filename, will merge with existing, overwriting conflicts
build   filename    build .hex file and save as filename
exit                quit program
quit                exit program
new                 clear slate, start new project
show                show existing entries

literalize rulenum  convert a rule to literals
addrule qualifier modifier priority
delrule priority
editrule (qualifier, modifier, priority) newValue

*/



typedef struct entry_s {
   unsigned long address;
   unsigned long data;
} entry_t;


typedef struct chip_s {
   entry_t * data;
   rules_t rules;
   unsigned long volume;
   unsigned long df;
   char * filename;
   unsigned long end;  // end of addresses
} chip_t;



Status_t chipInit    ( chip_t * this) ;
Status_t chipDestroy ( chip_t * this) ;
Status_t chipAdd     ( chip_t * this, unsigned long Address, unsigned long Data) ;
Status_t chipSave    ( chip_t * this, FILE * output) ;
Status_t chipShow    ( chip_t * this );
Status_t chipFind    ( chip_t * this, unsigned long Address, unsigned long * index);
Status_t chipDel     ( chip_t * this, unsigned long Address) ;
Status_t buildBuffer ( chip_t * this, unsigned char ** data) ;
Status_t chipBuild   ( chip_t * this, char * filename) ;
Status_t chipWalk    ( chip_t * this );

void     cmdFileName ( chip_t * Project, char * params) ;
void     cmdSave     ( chip_t * Project, char * params) ;
void     cmdLoad     ( chip_t * Project, char * params) ;
void     cmdHelp     ( ) ;
Status_t cmdDefault  ( chip_t * Project, char * params);
Status_t cmdSet      ( chip_t * Project, char * params);
Status_t cmdUnset    ( chip_t * Project, char * params);
Status_t cmdMemEnd   ( chip_t * Project, char * params);
void     cmdBuild    ( chip_t * Project, char * params);
void     cmdRuleAdd  ( chip_t * Project, char * params);
void     cmdRuleDel  ( chip_t * Project, char * params);
void     cmdView     ( chip_t * Project);
void     cmdRuleMove ( chip_t * Project, char * params);

Status_t buffer2ihex( unsigned char * data, char * filename, unsigned long size) ;
void     makeProject ( ) ;
long     getNumber   ( char * s, char ** l) ;
void     FF          ( char ** string) ;
Status_t test        ( void ) ;
int      doCmd       ( char * string, chip_t * Project) ;
int      getCommand  ( char ** commands, char ** in) ;
void     chomp       ( char * string) ;



int main(int argc, char** argv) {
  int retCode;
  
  if (isatty(STDIN_FILENO)) {   
    // terminal
    printf("Terminal mode.\n");
  } else {
    // pipe
    printf("Pipe mode.\n");
  }
  
  retCode = 0;
  printf(" This program helps build intel hex files \n");
  printf(" Written by Dan Williams 2009-2012 \n");
  printf(" Ver. %s \n", version);
  makeProject();
  
  return(retCode);
}




void makeProject() {

  chip_t * Project;  
  char *  string;
  char    reply[REPLYSZ];

  string  = NULL;
  Project = NULL;
  Project = realloc(Project, sizeof(chip_t));
  chipInit(Project);
  
  printf(" Type \"help\" for help. \n");
  do {      
    if (&string) free(string);
    string =  readline("> ");
    add_history (string);            
  } while((string != NULL) && (!doCmd(string, Project))); 
    
   chipDestroy(Project); 
   free(Project);
}





Status_t chipInit(chip_t * this) {

  this->volume   = 0;
  this->df       = 0;
  this->end      = 0;
  this->data     = NULL;
  this->filename = NULL;
  initRules ( &(this->rules) ); 
  return OK;
  
}



Status_t chipDestroy(chip_t * this) {

  finiRules ( &(this->rules) ) ;
  free(this->data);
  
  return OK;
}



Status_t chipAdd(chip_t * this, unsigned long Address, unsigned long Data) {

  entry_t * temp;
  unsigned long result;
  
  if (chipFind(this, Address, &result) == OK) {
    this->data[result].data    = Data;
    return OK;
  }
  
  temp = this->data;
  this->volume++;
  
  this->data = realloc( this->data, sizeof(entry_t) * this->volume);
  
  this->data[this->volume-1].address = Address;
  this->data[this->volume-1].data    = Data;
  
  if (Address > this->end) this->end = Address;
  
  return OK;
}


Status_t chipDel(chip_t * this, unsigned long Address) {

  entry_t * temp;
  unsigned long i;
  unsigned long result;
  
  if (chipFind(this, Address, &result) != OK) return NoMatch;
  for (i = result; i < (this->volume)-1; i++) {    
   this->data[i].address = this->data[i+1].address;
   this->data[i].data    = this->data[i+1].data;
  }
  
  this->volume--;
  
  this->data = realloc( this->data, sizeof(entry_t) * this->volume);
  
  return OK;

}

Status_t chipFind(chip_t * this, unsigned long Address, unsigned long * index) {

  unsigned long i;
  
  for ( i = 0; i < this->volume; i++) {
    if (this->data[i].address == Address) {
      *index = i;
      return OK;
    }
  }
    
  return NoMatch;
}



Status_t chipSave(chip_t * this, FILE * output) {

   long i;

   fprintf(output, "default 0x%04X\n", this->df);

   // dump table
   for (i = 0; i < this->volume; i++) {
     fprintf(output, "set 0x%04X 0x%04X\n", this->data[i].address, this->data[i].data);
   }
   
   // dump rules
     for(i = 0; i < this->rules.count; i++) {
         fprintf(output, "addrule %s %s %d\n",this->rules.rules[i].qualifier, this->rules.rules[i].modifier, this->rules.rules[i].priority);
     }

   fprintf(output, "memend 0x%04X\n", this->end);
      
   return OK;
}



Status_t chipShow(chip_t * this) {
   long p, i;
   long rule;
   char buff[250];
   
   printf( "filename %s \n", this->filename);
   printf( "default 0x%04X\n", this->df);
   printf( "memory ends at 0x%04X\n", this->end);
   printf(" Address | Data \n");
   printf("------------------\n");
   for (i = 0; i < this->volume; i++) {
     printf( "0x%04X  -> 0x%04X\n", this->data[i].address, this->data[i].data);
   }
   printf("\n");
   printf("  Priority | Qualifier                  | Modifier         \n");
   printf("---------------------------------------------------------\n");
   
   //DumpRules ( &(this->rules) ) ;
   //printf("Starting at rule: %d\n", (this->rules.priTop)+2);
   
   p = (this->rules.priTop)+2;  
   
   for( ; nextRule( &(this->rules), &p, &i ) == OK;) {
     // printf("%ld, if %s then %s\n", this->rules.rules[i].priority, this->rules.rules[i].qualifier, this->rules.rules[i].modifier);
      //printf("rule %d: \n", i);
      snprintf(buff,    200-0 ,"  %ld                          ", this->rules.rules[i].priority );
      snprintf(buff+13, 200-13, "%s                            ", this->rules.rules[i].qualifier);
      snprintf(buff+42, 200-42, "%s                            ", this->rules.rules[i].modifier);
      printf("%s\n", buff);
   }
   
  return OK;
}

Status_t chipWalk(chip_t * this) {

  int size,  j;
  unsigned long i;
  unsigned char * data;
  long p, n;
  varlist_t vars;

  
  initVars(&vars);
  data = malloc(this->end);
  memset(data, this->df, this->end);  // default first
 // printf("Starting build\n");
  for (i = 0; i < this->end; i++) {                  // apply pre-literal rules (infinity down to 0)
    
    setVar     (&vars, "D", data[i], 1) ;
    setVar     (&vars, "A", i, 0) ;
    
    printf("Address %d:\n", i);
    for(p = (this->rules.priTop)+1; nextRule( &(this->rules), &p, &n ) == OK ;   ) {
    
    //  printf( "ph1: %s %s %d\n", this->rules.rules[n].qualifier, this->rules.rules[n].modifier, i);
     // printf("rule priority %d\n", p);
   
      if (p > 0)                                                       // filter out only rules above zero
        if(eq_Eval(this->rules.rules[n].qualifier, &vars) != (long)0) {            
          printf( "  Applying rule %d: q:%s  m:%s\n", p, this->rules.rules[n].qualifier, this->rules.rules[n].modifier);
          eq_Eval(this->rules.rules[n].modifier, &vars);        
          
        }
    }
    data[i] = eq_Eval("D", &vars);
    printf("  Resulting value: %d\n", data[i]); 
    
  }
  
  
  for (i = 0; i < this->volume; i++) {                  // apply literals
   //  printf( "ph2: %d\n", i);
     data[this->data[i].address] = this->data[i].data;
  }

  for (i = 0; i < this->end; i++) {                  // apply post-literal rules (infinity down to 0)
  //  printf( "ph3: %d\n", i);
    setVar     (&vars, "D", data[i], 1) ;
    setVar     (&vars, "A", i, 0) ;
    printf("Address %d:\n", i);
    for(p = 0;  nextRule( &(this->rules), &p, &n ) == OK;) {
     // printf(">> rule priority %d\n", p);
      if(eq_Eval(this->rules.rules[n].qualifier, &vars) != (long)0) {  
        printf( "  Applying rule %d: q:%s  m:%s\n", p, this->rules.rules[n].qualifier, this->rules.rules[n].modifier);
        eq_Eval(this->rules.rules[n].modifier, &vars);        
      }
    }
    
    data[i] = eq_Eval("D", &vars);
    printf("  Resulting value: %d\n", data[i]); 
  }  
  
  free(data);
  
  return OK;
}

Status_t buildBuffer(chip_t * this, unsigned char ** data) {
  int size,  j;
  unsigned long i;
  long p, n;
  varlist_t vars;

  
  initVars(&vars);
  *data = malloc(this->end);
  memset(*data, this->df, this->end);  // default first
 // printf("Starting build\n");
  for (i = 0; i < this->end; i++) {                  // apply pre-literal rules (infinity down to 0)
    
    setVar     (&vars, "D", (*data)[i], 1) ;
    setVar     (&vars, "A", i, 0) ;
    
    for(p = (this->rules.priTop)+1; nextRule( &(this->rules), &p, &n ) == OK ;   ) {
    
    //  printf( "ph1: %s %s %d\n", this->rules.rules[n].qualifier, this->rules.rules[n].modifier, i);
     // printf("rule priority %d\n", p);
   
      if (p > 0)                                                       // filter out only rules above zero
        if(eq_Eval(this->rules.rules[n].qualifier, &vars) != (long)0) {   
          eq_Eval(this->rules.rules[n].modifier, &vars);        
        }
    }
    
    (*data)[i] = eq_Eval("D", &vars);
  }
  
  
  for (i = 0; i < this->volume; i++) {                  // apply literals
   //  printf( "ph2: %d\n", i);
     (*data)[this->data[i].address] = this->data[i].data;
  }

  for (i = 0; i < this->end; i++) {                  // apply post-literal rules (infinity down to 0)
  //  printf( "ph3: %d\n", i);
    setVar     (&vars, "D", (*data)[i], 1) ;
    setVar     (&vars, "A", i, 0) ;
    
    for(p = 0;  nextRule( &(this->rules), &p, &n ) == OK;) {
     // printf(">> rule priority %d\n", p);
      if(eq_Eval(this->rules.rules[n].qualifier, &vars) != (long)0) {  
        eq_Eval(this->rules.rules[n].modifier, &vars);        
      }
    }
    
    (*data)[i] = eq_Eval("D", &vars);
  }
  
  return OK;
}

Status_t buffer2ihex( unsigned char * data, char * filename, unsigned long size) {
  inteldata ihex;
  int   j;
  unsigned long i;


  intelhex_fileopen ( &ihex, filename, 0); //!!!???!!! errors people, handle errors
  ihex.reclen       = 0x10;
  
  for (i=0; i < size; i++) {
    j = i % 0x10;
    ihex.recdata[j] = data[i];
    if ((j == 0x0F)) {
      ihex.offset = (i/0x10)*0x10;
      intelhex_writeline (&ihex);
    }    
  }  
  
  if ((j = i % 0x10) != 0) {
    ihex.reclen       = j;  
    ihex.offset = (i/0x10)*0x10;
    intelhex_writeline (&ihex);
  }
  
  intelhex_writeeof (&ihex);
  intelhex_fileclose (&ihex);
  
  return OK;
}


int doCmd(char * string, chip_t * Project) {

// hey doc, these gotta match!
  enum cmdnums { 
         EMPTY = 0,
         HELP,
         BUILD,
         QUIT,
         EXIT,
         SHOW,
         FILENAME,
         SAVE,
         LOAD,
         NEW,
         DEFAULT,
         SET,
         MEMEND,
         UNSET,
	 ADDRULE,
	 DELRULE,
         MOVERULE,
       EDITRULE,
         WALK,
         VIEW,
         UNKNOWN
         
  };  
  
  static char * commands[] = { 
         " ",
         "help", 
         "build", 
         "quit",
         "exit",
         "show",        
         "filename",
         "save",
         "load",
         "new",
         "default",
         "set",
         "memend",
         "unset",
	 "addrule",
	 "delrule",
         "moverule",
       "editrule",
         "walk",
         "view",
         "" 
  };
  
    chomp(string); 
       
    switch( getCommand(commands, &string)) {
      case HELP:
         cmdHelp();
         break;
      case BUILD:
         cmdBuild(Project, string);
         break;
      case EXIT:          
      case QUIT:
         return 1;
         break;
      case SHOW:
         chipShow(Project);
         break;
      case FILENAME:
         cmdFileName(Project, string);
         break;
      case SAVE:
         cmdSave(Project, string);
         break;
      case LOAD:
         cmdLoad(Project, string);
         break;    
      case NEW:    
         chipDestroy(Project);
         chipInit(Project);    
         break;  
      case DEFAULT:
         cmdDefault(Project, string);
         break;
      case SET:
         cmdSet(Project, string);
         break;
      case UNSET:
         cmdUnset(Project, string);
         break;
      case MEMEND:
         cmdMemEnd(Project, string);
         break;
      case ADDRULE:
         cmdRuleAdd(Project, string);
         break;
      case DELRULE: // delete rule by priority
         cmdRuleDel(Project, string);
         break;
      case MOVERULE: // change a rules priority
         cmdRuleMove(Project, string);
         break;
      case EDITRULE: // change qualifier and or modifier
         break;
      case WALK: // step thru compilation
         chipWalk( Project );
         break;
      case VIEW: // show resulting data table
         cmdView(Project);
         break;
      case UNKNOWN:
         printf("unrecognized command.\n");
         break;
    
    }
    
    return 0;

}


/*****************************************
    determine which command was entered
*****************************************/

int getCommand(char ** commands, char ** in) {

  char * cursor;
  char * sep = " \t";
  int  i; 

 while( (**in == ' ') || (**in == '\t')) (*in)++;  // skip white space 
 if (**in == '\0') return 0;               // 0 is EMPTY
 cursor = strsep(in, sep); 
 for( i = 0; ((strcmp(cursor, commands[i]) !=0 ) && (*commands[i] != '\0')) ; i++); 
 return i;
 
}



/***************************************
  add data rule
***************************************/
void cmdRuleAdd( chip_t * Project, char * params) {

  char * qual, * mod, * pri;
  long  n_pri;   
  char * sep = " \t";
   
  if ((params==NULL) || (strlen(params) == 0)) {
    printf("need Qualifier Modifier Priority. example   addrule 1 d=~d 10 \n");
    return;
  } else {  
    
    // tear off the 3 paramiters
   // pri  = strsep(&params, sep); // priority  
    qual = strsep(&params, sep); // qualifier
    mod  = strsep(&params, sep); // modifier
    pri  = strsep(&params, sep); // priority  
    
    if ((qual==NULL) || (mod==NULL) || (pri == NULL)) {
      printf("need Qualifier Modifier Priority. example   addrule 1 d=~d 10 \n");
      return;
    }
    
    n_pri = atol(pri);
    
    // rules with duplicate priorities will be overwritten
    createRule(&(Project->rules), qual, mod, n_pri);

   // printf( "%d: if %s then %s\n", n_pri , qual, mod);     
    //DumpRules(&(Project->rules));                
  } 
}


/***************************************
  add data rule
***************************************/
void cmdRuleMove( chip_t * Project, char * params) {

  char * from_pri, * to_pri;
  long  n_from, n_to;   
  char * sep = " \t";
  unsigned long n;
   
  if ((params==NULL) || (strlen(params) == 0)) {
    printf("need old and new priorities. example   moverule -10 -20 \n");
    return;
  } else {  
    
    // tear off the paramiters
    from_pri  = strsep(&params, sep); // existing priority
    to_pri    = strsep(&params, sep); // new priority
    
    n_from = atol(from_pri);
    n_to   = atol(to_pri);
    
    if (getRule(&(Project->rules), NULL, &n, n_from) == OK) {
      Project->rules.rules[n].priority  = n_to;
    } else {
      printf("No rule with priority %ld found\n", n_from);
    }   
  } 
}




/***************************************
 *   del data rule
 *   ***************************************/
void cmdRuleDel( chip_t * Project, char * params) {

  char * pri;
  unsigned long  n_pri;
  char * sep = " \t";

  if ((params==NULL) || (strlen(params) == 0)) {
    printf("need Priority to be deleted. example  delrule 10 \n");
    return;
  } else {
    pri  = strsep(&params, sep); // priority
    n_pri = atol(pri);
    deleteRule(&(Project->rules), n_pri);
    //printf( "%d: if %s then %s\n", n_pri , qual, mod);
   // DumpRules(&(Project->rules));
  }
}


/***************************************
  set/get project filename
***************************************/
void cmdFileName( chip_t * Project, char * params) {
  char * cursor;   
  char * sep = " \t";

  debug("entering cmdfilename");
 
  if ((params==NULL) || (strlen(params) == 0)) {
    if (Project->filename != NULL) 
       printf("%s\n", Project->filename);
    return;
  } else {  
    if (Project->filename != params) { // we tend to get called with our own pointer
      if (Project->filename != NULL) {
        free(Project->filename);      
      } 
      Project->filename = strdup(params);
    }
  } 
  
}


/*************************************
    save machine to file
************************************/
void cmdSave( chip_t * Project, char * params) {
   char * cursor;   
   char * sep = " \t";
   FILE * output;
   unsigned int i, j;
   
  debug("entering cmdsave"); 
  if ((params==NULL) || (strlen(params) == 0) ) {
     if (Project->filename != NULL) {
       debug("calling self with filename");
       cmdSave( Project, Project->filename);
     } else {
       printf ("I need you to give me a filename!\n");
     }
     return;
  } else {   
    debug("seperating commands");
    while( cursor = strsep(&params, sep) ) { 
      if ((output = fopen(cursor, "wb")) == NULL) {
        printf("Couldn't open %s to save machine.\n", cursor);
      } else {
        
        chipSave(Project, output);
        
        debug("closing file");
        fclose(output);
        debug("resetting filename");
        cmdFileName(Project, cursor);
      }                   
    }
  }           
}



/**********************************
  build an ihex file and save it
**********************************/
void cmdBuild( chip_t * Project, char * params) {
   char * cursor;
   char * filename;   
   char * sep = " \t";
   FILE * output;
   unsigned int i, j;
   unsigned char * data;
   
  debug("entering cmdbuild"); 
  if ((params==NULL) || (strlen(params) == 0) ) {
     if (Project->filename != NULL) {
       debug("calling self with filename");
       cmdBuild( Project, Project->filename);
     } else {
       printf ("I need you to give me a filename!\n");
     }
     return;
  } else {   
    debug("seperating commands");
    buildBuffer(Project, &data);   // build image
    
    while( cursor = strsep(&params, sep) ) { 
      filename = strdup(cursor);
      filename = realloc(filename, strlen(filename)+6); //!!!???!!!      
      strcat(filename, ".hex");
        
      //chipBuild(Project, filename);
      buffer2ihex( data, filename, Project->end);

      free(filename);
      debug("resetting filename");
      cmdFileName(Project, cursor);
    }
    free(data);
  }           
}



/**********************************
  view data generated
**********************************/
void cmdView( chip_t * Project) {
  
  unsigned char * data;
  int count;
  long int countdown;
  unsigned int countup;
  int got;
  
  
  countdown = Project->end;
  countup = 0;
  
  buildBuffer(Project, &data);
   
   while((got = Limit(countdown, 0, 16)) > 0 ) {
   
     // data conversions...
        printf("%04lX:", countup);  // print address
        for (count = 0; count < got; count++) {      
         if ((count == 8)) {  printf(" -"); }         
         printf(" %02X", data[count+countup]);                 
        }
        // character translations for old school dump format  
        printf("  ");
        for (count = 0; count < got; count++) {
          if ((data[count+countup] >= 32) && (data[count+countup] < 127)) {
            printf("%c", data[count+countup]);
          } else {
            printf(".");
          }        
        }              
    countdown -= 16;
    countup += 16;
    printf("\n");
  }     

  free(data);
 
}

/**********************************
  load file into current machine
**********************************/

void cmdLoad( chip_t * Project, char * params) {
   char * cursor;   
   char * sep = " \t";
   FILE * input;
   char * string;
   char reply[REPLYSZ];
   
  if ((params==NULL) || (strlen(params) == 0)) {
    if (Project->filename != NULL) {
       cmdLoad( Project, Project->filename);
     } else {
       printf ("I need you to give me a filename!\n");
     }
     return;
  } else {   
    while( cursor = strsep(&params, sep) ) { 
      if ((input = fopen(cursor, "rb")) == NULL) {
        printf("Couldn't open %s to load.\n", cursor);
      } else {      
        do {   
          string = fgets(reply, REPLYSZ, input);      
        } while((string != NULL) && (!doCmd(string, Project)));            
        fclose(input);
        cmdFileName(Project, cursor);
      }                   
    }    
  }           
}


/*********************************
  print command help
*********************************/

void cmdHelp() {

printf( "\
  Priority goes is done highest to lowest. at zero, the fixed   \n\
   values are applied. Values may be negitive.                  \n\
                                                                \n\
help                         print help (your looking at it)    \n\
show                         show project                       \n\
build                        make ihex file                     \n\
walk                         list rules as applied during build \n\
quit                         exit                               \n\
filename {filename}          report or set the project filename \n\
save {filename}              save  project                      \n\
load {filename}              load project                       \n\
new                          clear project                      \n\
default n                    set default memory value to n      \n\
set a d                      set or reset address to data       \n\
 values can be decimal (38) hex (0x38) or binary (0b00111000)   \n\
unset a                      delete an entry                    \n\
memend n                     specify the size of the table      \n\
addrule q m p                add rule with Qualifer, Modifer    \n\
                               and Priority                     \n\
delrule p                    delete rule at Priority            \n\
moverule p1 p2               change priority of rule from       \n\
                               p1 to p2                         \n\
editrule q m p               change rule priority p to          \n\
                               qualifier q and modifier m       \n\
                                                                \n\
view                         preview output                     \n\
                                                                \n\
           {} denote optional paramiters.                       \n\
");
}



Status_t cmdDefault( chip_t * Project, char * params){

    if ((params==NULL) || (strlen(params) == 0)) {
       printf("0x%04X\n", Project->df);
    return OK;
  } else {   
      Project->df = getNumber(params, NULL);
  } 

}


Status_t cmdMemEnd( chip_t * Project, char * params){

    if ((params==NULL) || (strlen(params) == 0)) {
       printf("0x%04X\n", Project->end);
    return OK;
  } else {   
      Project->end = getNumber(params, NULL);
  } 

}


Status_t cmdUnset( chip_t * Project, char * params){
    unsigned long address;

    if ((params==NULL) || (strlen(params) == 0)) {
       printf("I need an address. \n");
       return BadArg;
    } else {        
        address = getNumber(params, NULL);        
        chipDel( Project, address);                     
  } 
    return OK;
}


Status_t cmdSet( chip_t * Project, char * params){
    char * marker;
    unsigned long address, data;

    if ((params==NULL) || (strlen(params) == 0)) {
       printf("I need an address and data. \n");
       return OK;
    } else {  
      
        address = getNumber(params, &marker);
        data    = getNumber(marker, NULL);
        
        chipAdd( Project, address, data);
                     
  } 

}
   
/**************************************
   tear off any trailing linefeeds
**************************************/

void chomp(char * string) {
  char * target;
  
  target = strchr(string,'\n');
  if (target != NULL) { *target = '\0'; }
  
}     

/**************************************
  Forward pointer thru whitespace
***************************************/
void FF(char ** string) {
  while((**string == ' ') || (**string == '\t')) (*string)++;
}
         
     
/**************************************
  Get a number in decimal, octal, hex, or binary
***************************************/    
long getNumber(char * s, char ** l) {
  char * temp;
  temp = s;
  FF(&temp);
  if (strncmp(temp, "0b", 2)==0) { // pick off binary
    return strtol(temp+2, l, 2);
  } else {
    return strtol(temp, l, 0);    
  }         
}         
         
         
         
         
