
#include "rules.h"



/*
 return the priority and index 
 of the rule with highest priority lower than pri
 
  this  is our source data
  pri   is the priority to work from
  index is the returned rule index number
*/
Status_t nextRule( rules_t * this, long * pri, long * index) {
 
  long retpri; // return priority
  long reti;   // return index
  long n, p;   // itterator and priority

  reti   = -1;                       // invalid index
  retpri = (this->priBot) - 1;         // lowest priority avilable
  
  for(n = 0; n < this->count; n++) { // for all rules...
    p = this->rules[n].priority;     // get the priority...
    //printf("p: %d, < *pri %d,   > retpri: %d\n", p, *pri, retpri);
    
    if ( (p < *pri) && (p > retpri) ) { // highest number greater than 'last'
      retpri = p;  reti = n;          // remember it
    }
    
  }
  
  if (reti == (-1)) { // nothing was found
    return NoMatch;
  } else {            
    *pri   = retpri;
    *index = reti;
    return OK;
  }
  
}


Status_t initRules( rules_t * this ) {
  this->rules  = NULL;
  this->count  = 0;
  this->priTop = 0;
  this->priBot = 0;
  return OK;
}
 
 
Status_t finiRules( rules_t * this ) {

  unsigned long n;

  for(n = 0; n < this->count; n++) {
    free(this->rules[n].qualifier);
    free(this->rules[n].modifier);
  }
  if(this->rules) free(this->rules);
  this->count = 0;
  return OK;
}
 
 
Status_t ShowRule( rule_t * this ) {
  printf(" qualifier: %s\n", this->qualifier);
  printf(" modifier:  %s\n", this->modifier);
  printf(" priority:  %d\n", this->priority);
  return OK;
} 
 
 
Status_t DumpRules( rules_t * this ) {

  unsigned long n;
  
  printf("total rules: %d\n", this->count );
  
  for(n = 0; n < this->count; n++) {
    printf("%d:\n", n);
    ShowRule(&this->rules[n]);
  }
  
  printf("Top priority: %ld   Bottom priority: %ld\n", this->priTop, this->priBot);
  
  return OK;
}
 
 
Status_t createRule( rules_t * this, char * qualifier, char * modifier, long int priority ) {
 
  unsigned long n;
 
  if (getRule(this, NULL, &n, priority) == OK) {
    
    this->rules[n].qualifier = strdup(qualifier);
    this->rules[n].modifier  = strdup(modifier);
    this->rules[n].priority  = priority;
    
  } else {

    this->rules = realloc(this->rules, sizeof(rule_t)*((this->count)+1));

    this->rules[this->count].qualifier = strdup(qualifier);
    this->rules[this->count].modifier  = strdup(modifier);
    this->rules[this->count].priority  = priority;
    this->count++;

    if (this->priTop < priority) {
      this->priTop = priority;
    } else if (this->priBot > priority) {
      this->priBot = priority;
    }

  }
  return OK; 
}
 
 
 
 
Status_t getRule( rules_t * this, rule_t ** result, unsigned long * index, long int priority ) {
 // find it
 
  unsigned long n;
  
  for(n = 0; n < this->count; n++) {
    if( this->rules[n].priority == priority) {
      if (result) { *result = &this->rules[n]; }
      if (index)  { *index  = n; }
      return OK;
    }
  }
  
  return NoMatch;
}
 
 
Status_t deleteRule( rules_t * this, long int priority ) {
  
  unsigned long n;
  
 // free elements
 
 // n = priority;
  if (getRule(this, NULL, &n, priority) != OK) {
    return NoMatch;
  }
  
  free(this->rules[n].qualifier);
  free(this->rules[n].modifier);

 // fill gap with last item
  this->rules[n].qualifier = this->rules[this->count-1].qualifier;
  this->rules[n].modifier  = this->rules[this->count-1].modifier;
  this->rules[n].priority  = this->rules[this->count-1].priority;

 // delete last item
  
  this->rules = realloc(this->rules, sizeof(rule_t)*(this->count-1));
  this->count--;
  return OK;
}
 
 
/* 
 
int main(void) {
  rules_t myrules;
  rule_t * myrule;
  long  temp;

  initRules(& myrules); 
  
  createRule( &myrules, "qual",  "mod",   20);
  createRule( &myrules, "qual2", "mod2", -10);
  createRule( &myrules, "qual3", "mod3",  10);
  
  nextRule( &myRules, 5, &temp);

  //DumpRules( &myrules );
  
  deleteRule( &myrules, -10);
  
  if (getRule( &myrules, & myrule, NULL, 10) == OK) {
   printf("Found:\n");
   ShowRule( myrule );
  } else {
   printf("Search failed\n");
  }
  
  DumpRules( &myrules ) ;
 
  finiRules( & myrules);
  return 0; 
}

*/

